/*
 *      getcompression.c
 *
 *      Copyright 2011 Anshul Narula aka sunny <sunny_delhi18@yahoo.com>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

/*bash builtin related includes*/
#include <config.h>
#if defined (HAVE_UNISTD_H)
#  include <unistd.h>
#endif
#include "bashansi.h"
#include "builtins.h"
#include "shell.h"
#include "bashgetopt.h"

/*Get compression program used for a package using magic number similar to what
 * newer tar does using their automatic detection of compression. In part, this
 * code is influenced by how tar-1.26 does it*/

/*These are the program names to be used for uncompressing*/
#define ZENPAKT_PKG_GZIP_PROG  "gzip"
#define ZENPAKT_PKG_BZIP2_PROG "bzip2"
#define ZENPAKT_PKG_LZMA_PROG  "lzma"
#define ZENPAKT_PKG_XZ_PROG    "xz"

/*This structure stores everything about a compression*/
typedef struct zenpakt_pkg_cmagic {
  size_t length;
  char *magic;
  char *program;
  char *extension;
} ZENPAKT_PKG_CMAGIC;

/*This has been altered to change the magic for lzma using what I found in case of
 * file(1) in /etc/file/magic/compress, the magic used with tar-1.26 doesnot work */
ZENPAKT_PKG_CMAGIC zenpakt_pkg_cmagictab[] = {
  { 2, "\037\213",             ZENPAKT_PKG_GZIP_PROG,  "tgz"  },
  { 3, "BZh",                  ZENPAKT_PKG_BZIP2_PROG, "tbz"  },
  { 4, "\x5d\x00\x00\x80",     ZENPAKT_PKG_LZMA_PROG,  "tlz"  },
  { 6, "\xFD" "7zXZ",          ZENPAKT_PKG_XZ_PROG,    "txz"  },
};

#define NMAGIC (sizeof(zenpakt_pkg_cmagictab)/sizeof(zenpakt_pkg_cmagictab[0]))

/*Check for compression by comparing the magic for the file. If none matches or
 * file could not be fopened return NULL */
ZENPAKT_PKG_CMAGIC *zenpakt_pkg_get_compression(char *file) {
  ZENPAKT_PKG_CMAGIC *p, *ret=NULL;
  void *buf;
  FILE *stream;

  if ( (stream=fopen(file, "r")) == NULL ) {
    fprintf(stderr, "%s\n", strerror(errno));
    return NULL;
  }

  /*10 should be enough as the bytes we compare are not more than 6*/
  buf = malloc(10);

  fread(buf, 10, 1, stream);

  for (p = zenpakt_pkg_cmagictab ; p < zenpakt_pkg_cmagictab + NMAGIC; p++)
    if (memcmp (buf, p->magic, p->length) == 0) {
      ret=p;
      break;
    }

  fclose(stream);
  free(buf);
  return ret;
}

/*check wether extension is the correct one for the magic. You can use this to
 *check wether a magic returned for a file by zenpakt_pkg_get_compression
 *matches with the extension of the file correctly. We require you to
 * provide the extension, obtaining it from filename can easily be done
 *in most languages*/
int zenpakt_pkg_valid_extension(ZENPAKT_PKG_CMAGIC *magic, char *extension) {
  return strcmp(magic->extension, extension) == 0? 1 : 0 ;
}

int zenpakt_pkg_valid_program(ZENPAKT_PKG_CMAGIC *magic, char *program) {
  return strcmp(magic->program, program) == 0? 1 : 0 ;
}

/* zenpakt_pkg_cmagic [options] compressed_file
 * -e  check wether given extension matches the required extension
 * -E  store the found extension in associated var
 * -p  check wether progname matches the required progname
 * -P  store the found program name in the given var
 */
int zenpakt_pkg_cmagic_builtin(WORD_LIST *list) {
  int c, opt;
  char *ext=NULL, *prog=NULL, *ext_varname=NULL, *prog_varname=NULL ;
  ZENPAKT_PKG_CMAGIC *p;

  reset_internal_getopt ();
  while ((opt = internal_getopt (list, "e:E:p:P:")) != -1) {
    switch (opt) {

      case 'e': /*verify extension*/
        ext=list_optarg;
        break;

      case 'E':
        if (legal_identifier (list_optarg) || valid_array_reference (list_optarg))
          ext_varname=list_optarg;
        else {
          sh_invalidid (list_optarg);
          return EXECUTION_FAILURE;
        }
        break;

      case 'p': /*verify program*/
        prog=list_optarg;
        break;

      case 'P': /*store the compression program in given variable*/
        if (legal_identifier (list_optarg) || valid_array_reference (list_optarg))
          prog_varname=list_optarg;
        else {
          sh_invalidid (list_optarg);
          return EXECUTION_FAILURE;
        }
        break;

      default:
        builtin_usage();
        return(EX_USAGE);
    }
  }

  list = loptend;

  if (list == NULL) return EXECUTION_FAILURE;

  p=zenpakt_pkg_get_compression(list->word->word);

  if (p == NULL) return EXECUTION_FAILURE;

  if (ext && zenpakt_pkg_valid_extension(p, ext) == 0)
    return EXECUTION_FAILURE;

  if (prog && zenpakt_pkg_valid_program(p, prog) == 0)
    return EXECUTION_FAILURE;

  if (ext_varname) {
    if (valid_array_reference (ext_varname) == 0)
        bind_variable (ext_varname, p->extension, 0);
      else
        assign_array_element (ext_varname, p->extension, 0);
      stupidly_hack_special_variables (p->extension);
  }

  if (prog_varname) {
    if (valid_array_reference (prog_varname) == 0)
        bind_variable (prog_varname, p->program, 0);
      else
        assign_array_element (prog_varname, p->program, 0);
      stupidly_hack_special_variables (p->program);
  }

  return EXECUTION_SUCCESS;

}

char *zenpakt_pkg_cmagic_doc[] = {
	"zenpakt_pkg_cmagic [-e ext] [-E evar] [-p prog] [-P pvar] file",
	"",
	"check wether magic for FILE matches one of the supported compression format:",
  "gzip, bzip2, lzma and xz. If it matches none return non-zero. If -e was given,",
  "additionally check wether EXT corresponds to the compression format matched.",
  "If -E is used place the value of corresponding extension in EVAR",
  "If -p is used, check wether PROG matches the program name for the matched",
  "compression format. If -P is used, store the value of program name corresponding",
  "to the matched compression format.",
  "",
  "For each supported format the extension and the program name are given below:",
  "",
  "format  extension  program",
  "======  =========  =======",
  "lzma     tlz         lzma",
  "bzip2    tbz         bzip2",
  "gzip     tgz         gzip",
  "xz       txz         xz",
	(char *)NULL
};

struct builtin zenpakt_pkg_cmagic_struct = {
	"zenpakt_pkg_cmagic",		/* builtin name */
	zenpakt_pkg_cmagic_builtin,		/* function implementing the builtin */
	BUILTIN_ENABLED,	/* initial flags for builtin */
	zenpakt_pkg_cmagic_doc,		/* array of long documentation strings. */
	"metaget filename",		/* usage synopsis; becomes short_doc */
	0			/* reserved for internal use */
};

#if 0
int main(int argc, char **argv) {
  ZENPAKT_PKG_CMAGIC *p;
  p=zenpakt_pkg_get_compression(argv[1]);

  if (p == NULL)
    return 1;

  if (zenpakt_pkg_valid_extension(p, argv[2]))
    printf("%s\n", p->program);
  else {
    fprintf(stderr, "extension %s doesnot match detected compression %s\n", argv[2], p->program);
    return 1;
  }

  fflush(stdout);
  fflush(stderr);
  return 0;
}
#endif
